@@ -21,10 +21,29 @@
 /// Any sensor-related data quoted in this documentation are orientative and are 
 /// taken from Version 2 - July 2021 of the datasheet
 ///
+/// ## Features
+///
+/// - "fp": enable this feature if your CPU or MCU has a floating-point unit
+///   and you want to be able to convert measured values to SI units (degrees
+///   Celsius, degreed Fahrenheit for temperature and % for relative humidity).
+///
+///   The driver stores measurements in milli degrees Celsius or milli degrees
+///   Fahrenheit for temperature and in per cent mille (pcm) for relative 
+///   humidity, and does conversions using fixed-point arithmetic to minimise
+///   loss of precision. Therefore, a temperature of 23.89 degrees Celsius will
+///   be stored as 23890 and and a humidity of 56.2% is stored as 56200.
+///
+///   The fp feature adds data structures and a function for converting a
+///   Measurement structure to a SIMeasurement structure, converting to floating
+///   point as the very last step (after any other calibration offsets have
+///   been applied to the measured values).
+///
 /// ## Usage:
 ///
 /// Import this crate and an `embedded_hal` implementation, then instantiate the
-/// driver; For example, assuming you have connected a SH40 sensor to a linux
+/// driver; 
+///
+/// * For example, assuming you have connected a SH40 sensor to a linux
 /// machine and it is detected as an i2c device:
 ///
 /// ```no_run
@@ -39,13 +58,45 @@
 ///     let mut sht40 = SHT40Driver::new(i2c_dev, I2CAddr::SHT4x_A, Delay);
 ///     
 ///     if let Ok(measurement) =
-///         sht40.get_temp_and_rh(Precision::High, TempUnit::Celsius) {
+///         sht40.get_temp_and_rh(Precision::High, TempUnit::MilliDegreesCelsius) {
 ///       println!("Temp: {temp} C, Relative Humidity: {rh} %",
 ///                temp = measurement.temp,
-///                rh = measurement.rel_hum_percent);
+///                rh = measurement.rel_hum_pcm);
 ///     }
 /// }
 /// ```
+///
+/// * Likewise, if you are using an ESP32-based board:
+///
+/// ```ignore
+/// use esp_idf_sys;
+/// use esp_idf_hal::delay;
+/// use esp_idf_hal::i2c;
+/// use esp_idf_hal::prelude::*;
+/// use sensor_temp_humidity_sht40::{SHT40Driver, I2CAddr, Precision, TempUnit};
+///
+/// fn main() {
+///     let peripherals = Peripherals::take().unwrap();
+///     let pins = peripherals.pins;
+///
+///     let i2c = peripherals.i2c0;
+///     let scl = pins.gpio9;
+///     let sda = pins.gpio8;
+///     let i2c_config = <i2c::config::MasterConfig as Default>::default()
+///                      .baudrate(400.kHz().into());
+///
+///     let mut sensor_drv = SHT40Driver::new(
+///         i2c::Master::<i2c::I2C0, _, _>::new(i2c, 
+///                                             i2c::MasterPins { sda, scl }, 
+///                                             i2c_config).unwrap(), 
+///         I2CAddr::SHT4x_A, 
+///         delay::Ets);
+///
+///     let measurement = sensor_drv.get_temp_and_rh(Precision::High,
+///                                                  TempUnit::MilliDegreesCelsius);
+///     println!("Measurement {:#?}", measurement);
+/// }
+/// ```
 
 use embedded_hal as hal;
 
@@ -54,7 +105,11 @@ use hal::blocking::i2c::{Read, Write, WriteRead};
 
 use sensirion_i2c::{crc8, i2c};
 
-// I2C Commands supported by the SHT40 sensor
+pub const MAX_MILLICELSIUS_OFFSET:i16 = 17000;
+pub const MAX_MILLIFAHRENHEIT_OFFSET:i16 = 32000;
+pub const MAX_PCMHUM_OFFSET:i16 = 32000;
+
+/// I2C Commands supported by the SHT40 driver
 #[allow(non_camel_case_types)]
 #[derive(Clone, Copy)]
 enum Command {
@@ -105,6 +160,7 @@ pub enum Precision {
     Low
 }
 
+/// Define the power at which the built-in heater should be operated.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum HeaterPower {
     /// High is typically 200mW (for VDD=3.3V)
@@ -117,6 +173,9 @@ pub enum HeaterPower {
     Low,
 }
 
+/// Define the duration for which the built-in heater should be turned on. After
+/// that time, the heater automatically turns off, without the need for any
+/// additional I2C commands.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum HeaterDuration {
     /// PulseLong is typically 1s (max 1.1s)
@@ -150,10 +209,37 @@ pub enum I2CAddr {
 /// unit transformation which might reduce accuracy.
 #[derive(Debug, PartialEq, Eq, Clone, Copy)]
 pub enum TempUnit {
-    Farrenheit,
+    MilliDegreesFahrenheit,
+    MilliDegreesCelsius,
+}
+
+/// SI Measurement unit for Temperature.
+///
+/// Used only with `SIMeasurement` when the fp feature is enabled
+#[cfg(feature="fp")]
+#[derive(Debug, PartialEq, Eq, Clone, Copy)]
+pub enum SITempUnit {
+    Fahrenheit,
     Celsius,
 }
 
+/// Temperature offset to be applied to sensor measurement after conversion from
+/// raw value. This can be used to apply external calibration, beyond what
+/// is automatically done on-sensor.
+///
+/// Here, TempUnit refers to the unit of the offset (delta). If the measurement
+/// values are requested in a different TempUnit, this offset will be 
+/// automatically transformed so that it can be correctly applied to values in
+/// the different TempUnit.
+#[derive(Debug, Clone, Copy)]
+pub struct TempOffset(i16, TempUnit);
+
+/// Humidity PCM offset to be applied to sensor measurement after conversion
+/// from raw value. This can be used to apply external calibration, beyond what
+/// is automatically done on-sensor.
+#[derive(Debug, Clone, Copy)]
+pub struct HumPcmOffset(i16);
+
 /// SHT40Driver is the main structure with which a user of this driver
 /// interracts. It is generic over two parameters implementing embedded_hal 
 /// traits for the specific platform you're using: I2c which must implement 
@@ -192,8 +278,12 @@ pub struct SHT40Driver<I2c, Delay> {
     i2c: I2c,
     address: u8,
     delay: Delay,
+    temp_offset: Option<TempOffset>,
+    hum_pcm_offset: Option<HumPcmOffset>,
 }
 
+/// Driver error type parameterised to support hal-specific communication error
+/// subtypes
 #[derive(Debug)]
 pub enum Error<E> {
     /// Error on the I2C bus
@@ -204,6 +294,14 @@ pub enum Error<E> {
     DelayError,
 }
 
+/// Driver error type independent of hal-specific types. 
+#[derive(Debug)]
+pub enum SHT40Error {
+    /// The offset applied is too large to be supported. This is returned to
+    /// avoid fixed-point arithmetic overflow errors.
+    CalibrationOffsetTooLarge,
+}
+
 impl<E, I2cWrite, I2cRead> From<i2c::Error<I2cWrite, I2cRead>> for Error<E>
 where
     I2cWrite: Write<Error = E>,
@@ -218,7 +316,9 @@ where
     }
 }
 
-
+/// I2C command code and maximum duration required for the sensor to respond
+/// to that particular command (wait time). If a I2C read is issued before
+/// max_duration_ms, the sensor may respond with NACK.
 #[derive(Clone, Copy)]
 struct DeviceCommand {
     cmd_code: u8,
@@ -229,12 +329,32 @@ struct DeviceCommand {
 ///
 /// You obtain a measurement by calling get_temp_and_rh(...) or 
 /// set_heater_then_measure(...) on an instance of SHT40Driver.
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, PartialEq, Eq, Clone, Copy)]
 pub struct Measurement {
+    /// Measured temperature
+    pub temp: i32,
+    /// Unit of temperature measurement (milliC or milliF)
+    pub temp_unit: TempUnit,
+    /// Measured relative humidity (pcm = percent mille = per 100_000)
+    pub rel_hum_pcm: u32,
+    /// The precision that was requested when performing the measurement.
+    /// For measurements obtained through set_heater_then_measure(...),
+    /// the sensor automatically performs a Precision::High measurement.
+    pub precision: Precision,
+}
+
+/// A sensor SI measurement result.
+///
+/// This is obtained by calling `convert_measurement_to_si` with an existing
+/// Measurement object. Only available whem the "fp" feature is enabled as 
+/// temperature and relative humidity (%) are stored as floating point numbers.
+#[cfg(feature="fp")]
+#[derive(Debug, Clone, Copy)]
+pub struct SIMeasurement {
     /// Measured temperature
     pub temp: f32,
     /// Unit of temperature measurement (C or F)
-    pub temp_unit: TempUnit,
+    pub temp_unit: SITempUnit,
     /// Measured relative humidity (%)
     pub rel_hum_percent: f32,
     /// The precision that was requested when performing the measurement.
@@ -255,10 +375,6 @@ pub struct RawMeasurement {
     precision: Precision,
 }
 
-/// Structure for storing the sensor serial number
-#[derive(Debug, PartialEq, Eq, Clone, Copy)]
-pub struct SerialNumber(u16, u16);
-
 
 impl Command {
     fn as_device_command(self) -> DeviceCommand {
@@ -266,7 +382,7 @@ impl Command {
             Command::SoftReset => DeviceCommand { cmd_code: 0x94,
                                                   max_duration_ms: 1 },
             Command::GetSerialNumber => DeviceCommand { cmd_code: 0x89,
-                                                        max_duration_ms: 0 },
+                                                        max_duration_ms: 1 },
             Command::MeasureTempAndHumidity(prec) => {
                 match prec {
                     Precision::High => DeviceCommand { cmd_code: 0xFD,
@@ -328,9 +444,79 @@ where
     ///             the sensor to respond to a measurement request or for the
     ///             heater to be deactivated.
     pub fn new(i2c: I2C, address: I2CAddr, delay: D) -> Self {
-        SHT40Driver { i2c, address: address as u8, delay }
+        SHT40Driver { i2c, 
+                      address: address as u8, 
+                      delay, 
+                      temp_offset: None, 
+                      hum_pcm_offset: None }
     }
 
+    /// Set a temperature offset that is automatically added to the value 
+    /// measured by the sensor for measurements (not raw only).
+    ///
+    /// This call will fail if the absolute value of TempOffset is larger than 
+    /// a pre-defined constant, depending on the TempUnit:
+    ///
+    /// MAX_MILLICELSIUS_OFFSET for TempUnit::MilliDegreesCelsius and
+    /// MAX_MILLIFAHRENHEIT_OFFSET for TempUnit::MilliDegreesFahrenheit
+    ///
+    /// This is to avoid overlows during fixed-point arithmetic operations
+    /// involving the offset.
+    pub fn set_temp_offset(&mut self, offset: TempOffset) -> Result<(), SHT40Error> {
+        let too_large = match offset.1 {
+            TempUnit::MilliDegreesCelsius => { 
+                if offset.0.abs() > MAX_MILLICELSIUS_OFFSET { 
+                    true 
+                } else { 
+                    false 
+                } 
+            },
+            TempUnit::MilliDegreesFahrenheit => { 
+                if offset.0.abs() > MAX_MILLIFAHRENHEIT_OFFSET { 
+                    true 
+                } else { 
+                    false
+                } 
+            }
+        };
+        if !too_large {
+            self.temp_offset = Some(offset);
+            Ok(())
+        } else {
+            Err(SHT40Error::CalibrationOffsetTooLarge)
+        }
+    }
+
+    /// Clear any previously set temperature offset
+    pub fn clear_temp_offset(&mut self) {
+        self.temp_offset = None;
+    }
+
+    /// Set a humidity offset that is automatically added to the value 
+    /// measured by the sensor for measurements (not raw only).
+    ///
+    /// This call will fail if the absolute value of TempOffset is larger than 
+    /// a pre-defined constant, MAX_PCMHUM_OFFSET
+    ///
+    /// This is to avoid overlows during fixed-point arithmetic operations
+    /// involving the offset.
+    pub fn set_hum_offset(&mut self, offset: HumPcmOffset) -> Result<(), SHT40Error> {
+        if offset.0.abs() < MAX_PCMHUM_OFFSET {
+            self.hum_pcm_offset = Some(offset);
+            Ok(())
+        } else {
+            Err(SHT40Error::CalibrationOffsetTooLarge)
+        }
+    }
+    
+    /// Clear any previously set humidity offset
+    pub fn clear_hum_offset(&mut self) {
+        self.hum_pcm_offset = None;
+    }
+    
+    /// Utility function for parsing the raw I2C sensor output buffer
+    ///
+    /// [not part of public API]
     fn parse_sensor_raw_measurement(rx_bytes: [u8; 6], precision: Precision, check_crc: bool)
         -> Result<RawMeasurement, Error<E>> {
             let temp_ticks = ((rx_bytes[0] as u16) << 8)  + (rx_bytes[1] as u16);
@@ -343,36 +529,27 @@ where
             Ok(RawMeasurement { temp_ticks, rel_hum_ticks, precision })
         }
 
-    fn convert_raw_to_units(raw_meas: RawMeasurement, temp_unit: TempUnit) -> Measurement {
-        let mut rel_hum_percent = -6.0 + 125.0 * (raw_meas.rel_hum_ticks as f32)/65535.0;
 
-        // calibration may result in non-physical values which according to the
-        // datasheet are only of interest when comparing the distribution of results
-        // between multiple sensors:
-        if rel_hum_percent > 100.0 { rel_hum_percent = 100.0; }
-        if rel_hum_percent < 0.0 { rel_hum_percent = 0.0; }
-
-        let temp = match temp_unit {
-            TempUnit::Celsius => -45.0 + 175.0 * (raw_meas.temp_ticks as f32)/65535.0,
-            TempUnit::Farrenheit => -49.0 + 315.0 * (raw_meas.temp_ticks as f32)/65535.0
-        };
-
-        Measurement {
-            temp,
-            temp_unit,
-            rel_hum_percent,
-            precision: raw_meas.precision
-        }
-    }
-
-    fn parse_measurement(rx_bytes: [u8; 6], precision: Precision, temp_unit: TempUnit, check_crc: bool)
+    /// Utility function for parsing the raw I2C sensor output buffer and
+    /// converting the temperature in the requested unit.
+    ///
+    /// This also applies the given offsets if provided, assuming that 
+    /// temp_offset is already converted for addition to a temperature in the
+    /// same unit as temp_unit. The humidity offset is assumed to be given
+    /// in pcm (per cent mille).
+    ///
+    /// [not part of public API]
+    fn parse_measurement(rx_bytes: [u8; 6], precision: Precision, temp_unit: TempUnit,
+                         check_crc: bool, temp_offset: Option<i16>, hum_pcm_offset: Option<i16>)
         -> Result<Measurement, Error<E>> {
             let raw = SHT40Driver::<I2C, D>::parse_sensor_raw_measurement(rx_bytes,
                                                               precision,
                                                               check_crc)?;
-            Ok(SHT40Driver::<I2C, D>::convert_raw_to_units(raw, temp_unit))
+            Ok(convert_raw_to_units(raw, temp_unit, temp_offset, hum_pcm_offset))
     }
 
+    /// Issue I2C command, wait for the maximum time it might take the sensor
+    /// to respond to that particular command, then read the response buffer.
     fn i2c_command_and_response(&mut self, cmd: Command, rx_bytes: Option<&mut [u8]>)
         -> Result<(), Error<E>>{
         let dev_cmd = cmd.as_device_command();
@@ -389,6 +566,10 @@ where
     /// consult the sensor datasheet for transformations which can be applied
     /// to this raw data to get temperature and humidity values.
     ///
+    /// No offsets set using `set_temp_offset` or `set_hum_offset` are applied,
+    /// as it is assumed the user of the API fully deals with any calibration
+    /// required beyond the one done on the sensor itself
+    ///
     /// The driver blocks for the max time the measurement can take, depending
     /// on the requested Precision, waiting for the measurement results.
     ///
@@ -409,9 +590,85 @@ where
                                                             false)
     }
 
+
+    /// The calibration offset and the measurement units are provided 
+    /// independently by users of this API. For example, the offset might
+    /// be set in milli degrees Fahrenheit, but the measurement is done in
+    /// milli degrees Celsius.
+    ///
+    /// Although this is not necessarily recommended because the need of
+    /// a further fixed-precision conversion of the offset, the API supports
+    /// the scenario by detecting the mismatch and converting.
+    ///
+    /// This is not a classical conversion of the offset from milliFahrenheit
+    /// to milliCelsius, because we're interested in the finite difference
+    /// (in a mathematical sense). If C is the temperature in Celsius and f(C)
+    /// is the conversion in Fahrenheit, we have:
+    ///
+    /// ```ignore
+    ///
+    /// f(C) = 9/5 * C + 32
+    /// f(C + off) = 9/5 * (C + off) + 32
+    /// 
+    /// ```
+    ///
+    /// Therefore, for an offset of "off" we have
+    /// 
+    /// ```ignore
+    /// 
+    /// f(C + off) - f(C) = 9/5 * off =>
+    /// f(C + off) - f(C) = 1.8 * off
+    /// 
+    /// ```
+    ///
+    /// This means that applying the offset 'off' to the the Celsius temperature
+    /// is equivalent to applying the offset 9/5 * off to the Fahrenheit 
+    /// temperature
+    ///
+    /// The function below implements exactly that using fixed precision
+    /// arithmetic.
+    ///
+    fn convert_temp_offset(temp_off: Option<TempOffset>, to_unit: TempUnit, 
+                           hum_pcm_off: Option<HumPcmOffset>) 
+        -> (Option<i16>, Option<i16>) {
+        let temp_offset: Option<i16>;
+        let hum_offset: Option<i16>;
+        
+        if let Some(t_off) = temp_off {
+            if t_off.1 != to_unit {
+               temp_offset = match to_unit {
+                   TempUnit::MilliDegreesCelsius => {
+                       // t_off.0 in milliFahrenheit, convert to milliCelsius
+                       Some(((t_off.0 as i32) * 555 / 1000) as i16)
+                       
+                   },
+                   TempUnit::MilliDegreesFahrenheit => {
+                       // t_off.0 delta in milliCelsius, convert to 
+                       // milliFahrenheit offset
+                       Some(((t_off.0 as i32) * 18 / 10) as i16)
+                   }
+               }
+            } else {
+                temp_offset = Some(t_off.0);
+            }
+        } else {
+            temp_offset = None;
+        }
+        hum_offset = hum_pcm_off.map(|off|{ off.0 });
+        (temp_offset, hum_offset)
+    }
+
     /// Get the results of a measurement, with temperature measured in the 
-    /// requested units. Relative humidity is unitless and always expressed as
-    /// a percent (%)
+    /// requested units (milli degrees C or milli degrees F).
+    ///
+    /// Relative humidity is unitless and always reported as
+    /// pcm (percent mille = per 100_000).
+    ///
+    /// You can easily convert into degrees C, degrees F (for temp) and percent
+    /// (%) for relative humidity by dividing the returned values by 1000. They
+    /// are stored as milli-* so that all internal conversions can be done using
+    /// integer arithmetic, without loss of precision. That way, MCUs without
+    /// a floating-point unit are supported.
     ///
     /// The driver blocks for the max time the measurement can take, depending
     /// on the requested Precision, waiting for the measurement results.
@@ -423,14 +680,20 @@ where
     ///                 more power. Consult the Precision struct documentation
     ///                 for details.
     /// * `temp_unit` - The measurement unit to be used for temperature. You can
-    ///                 choose between TempUnit::Celsius and TempUnit::Farrenheit
+    ///                 choose between TempUnit::MilliDegreesCelsius and 
+    ///                 TempUnit::MilliDegreesFahrenheit
     ///
     pub fn get_temp_and_rh(&mut self, precision: Precision, temp_unit: TempUnit)
         -> Result<Measurement, Error<E>> {
         let mut rx_bytes = [0; 6];
         let cmd = Command::MeasureTempAndHumidity(precision);
         self.i2c_command_and_response(cmd, Some(&mut rx_bytes))?;
-        SHT40Driver::<I2C, D>::parse_measurement(rx_bytes, precision, temp_unit, false)
+        let (temp_offset, hum_offset) = 
+            SHT40Driver::<I2C, D>::convert_temp_offset(self.temp_offset, 
+                                                       temp_unit, 
+                                                       self.hum_pcm_offset);
+        SHT40Driver::<I2C, D>::parse_measurement(rx_bytes, precision, temp_unit,
+                                                 false, temp_offset, hum_offset)
     }
     
     /// Activate the SHT40 built-in heater with the specified power and duration
@@ -466,8 +729,8 @@ where
     ///            duration, the heater is automatically deactivated.
     /// * `temp_unit` - The measurement unit to be used for temperature in the
     ///                 measurement taken just before the heater is deactivated.
-    ///                 You can choose between TempUnit::Celsius and 
-    ///                 TempUnit::Farrenheit
+    ///                 You can choose between TempUnit::MilliDegreesCelsius and 
+    ///                 TempUnit::MilliDegreesFahrenheit
     ///
     pub fn set_heater_then_measure(&mut self, hpow: HeaterPower, hdur: HeaterDuration, temp_unit: TempUnit)
         -> Result<Measurement, Error<E>> {
@@ -475,7 +738,12 @@ where
         let precision = Precision::High;
         let cmd = Command::SetHeaterThenMeasureHighPrec(hpow, hdur);
         self.i2c_command_and_response(cmd, Some(&mut rx_bytes))?;
-        SHT40Driver::<I2C, D>::parse_measurement(rx_bytes, precision, temp_unit, false)
+        let (temp_offset, hum_offset) = 
+            SHT40Driver::<I2C, D>::convert_temp_offset(self.temp_offset, 
+                                                       temp_unit, 
+                                                       self.hum_pcm_offset);
+        SHT40Driver::<I2C, D>::parse_measurement(rx_bytes, precision, temp_unit,
+                                                 false, temp_offset, hum_offset)
     }
     
     /// Activate the SHT40 built-in heater with the specified power and duration
@@ -495,12 +763,17 @@ where
     ///
     /// Additionally, the temperature sensor can be affected by the 
     /// thermally-induced mechanical stress from the heater, ofsetting the
-    /// measured temperature. Likewise, any offsets which need to be applied
-    /// to the measured values are the responsibility of the user of this API.
+    /// measured temperature. If you want a more accurate temperature 
+    /// measurement, it is recommended you wait ~1s after calling this function
+    /// and then get another normal measurement (not activating the heater).
     ///
     /// For more details, consult the sensor's datasheet or additional Sensirion
     /// application notes.
     ///
+    /// No offsets set using `set_temp_offset` or `set_hum_offset` are applied,
+    /// as it is assumed the user of the API fully deals with any calibration
+    /// required beyond the one done on the sensor itself
+    ///
     /// The driver blocks for the max time the heater is active, depending on
     /// the requested HeaterDuration, in order to get the measurement results.
     ///
@@ -523,7 +796,7 @@ where
     }
 
     /// Get the SHT40 sensor serial number
-    pub fn get_serial(&mut self) -> Result<SerialNumber, Error<E>> {
+    pub fn get_serial(&mut self) -> Result<u32, Error<E>> {
         let mut rx_bytes = [0; 6];
         let cmd = Command::GetSerialNumber;
         self.i2c_command_and_response(cmd, Some(&mut rx_bytes))?;
@@ -531,7 +804,7 @@ where
         let serial_a = ((rx_bytes[0] as u16) << 8)  + (rx_bytes[1] as u16);
         let serial_b = ((rx_bytes[3] as u16) << 8)  + (rx_bytes[4] as u16);
 
-        Ok(SerialNumber(serial_a, serial_b))
+        Ok( ((serial_a as u32) << 16) + (serial_b as u32) )
     }
 
     /// Perform a soft reset of the sensor. After this command, the user of the
@@ -545,14 +818,119 @@ where
             .map_err(|err| { Error::I2c(err) })?;
         Ok(())
     }
-
-
 }
 
+
+///
+/// The SHT40 sensor uses a 16-bit ADC for converting the analog 
+/// measurements into data, which is then calibrated on-sensor and then 
+/// sent via I2C. In the raw measurement, temp_ticks and rel_hum_ticks are
+/// direct outputs after the calibration process and take values in the
+/// [0 .. 2^16 - 1] interval.
+///
+/// In order to convert the ticks into usable values, we map this interval
+/// onto the following intervals:
+///
+/// For temp. in Celsius: [-45 .. 130]
+/// For temp. in Fahrenheit: [-49 .. 266] (equivalent to the C interval)
+/// For humidity (in %): [-6 .. 119], with values below 0 and above 100 not
+///                                   having physical meaning.
+///
+/// Mathematically, this means that if a and b are the limits of the interval
+/// [a .. b] for a particular metric (temp in C, temp in F, humidity) then 
+/// the value of the metric, M is computed according to:
+///
+/// ```ignore
+///
+///     (b - a) * Mticks 
+/// M = ----------------- + a                                       (1)
+///          2^16 - 1
+/// ```
+///
+/// However, we will not directly implement this formula but optimise it
+/// for fixed-point algebra. Therefore, the Measurement will hold only
+/// integer values rather than floating-point ones, and instead of the
+/// typical measurement units (C, F or %) we will output values 1000-times
+/// greater (units in milliC, milliF, pcm). 
+///
+/// Concretely, a temperature of 23.5 C will be returned as 23500 milliC and
+/// a humidity of 45.64 % will be returned as 45640 pcm (part per 100_000)
+///
+/// We write (1) as:
+///
+/// ```ignore
+///
+///               (b - a) * 1000 * Mticks
+/// M * 1000 = ---------------------------- + (a * 1000)           (2)
+///                    (               1  )
+///              2^i * ( 2^(16-i) -  -----)
+///                    (              2^i )
+/// ```
+///
+/// and we search for i that minimizes the fractional part, implementing the
+/// first division by 2^i as a bit shifit and the pre-computing an integer
+/// for ((b - a) * 1000) / (2^(16-i) - 1/2^i)
+/// 
+/// User provided offsets for temp and humidity can be used when externally
+/// calibrating the sensor. It is assumed that the unit for temp_offset is
+/// the same as temp_unit and that hum_pcm_offset is expressed in per cent 
+/// mille (pcm, part per 100_000).
+///
+/// The relative humidity returned as part of the measurement result is set to
+/// 0 if negative and to 100_000 if greater than 100_000 pcm (100%)
+/// 
+pub fn convert_raw_to_units(raw_meas: RawMeasurement, temp_unit: TempUnit,
+                            temp_offset: Option<i16>, hum_pcm_offset: Option<i16>)
+    -> Measurement {
+
+    let temp_offset = temp_offset.unwrap_or(0);
+    let hum_pcm_offset = hum_pcm_offset.unwrap_or(0);
+
+    // compute (2) with a = -6, b = 119, i = 13
+    let mut rel_hum_pcm = ((15625 * (raw_meas.rel_hum_ticks as i32)) >> 13) - 6000;
+    rel_hum_pcm = rel_hum_pcm + (hum_pcm_offset as i32);
+
+    // calibration may result in non-physical values which according to the
+    // datasheet are only of interest when comparing the distribution of results
+    // between multiple sensors:
+    if rel_hum_pcm > 100000 { rel_hum_pcm = 100000; }
+    if rel_hum_pcm < 0 { rel_hum_pcm = 0; }
+
+    let temp = match temp_unit {
+        // compute (2) with a = -45, b = 130, i = 13
+        TempUnit::MilliDegreesCelsius =>  ((21875 * (raw_meas.temp_ticks as i32)) >> 13) - 45000,
+        // compute (2) with a = -49, b = 266, i = 14
+        TempUnit::MilliDegreesFahrenheit => ((78751 * (raw_meas.temp_ticks as i32)) >> 14) - 49000
+    };
+
+
+    Measurement {
+        temp: temp + (temp_offset as i32),
+        temp_unit,
+        rel_hum_pcm: rel_hum_pcm as u32,
+        precision: raw_meas.precision
+    }
+}
+
+/// Converts a Measurement to a SIMeasurement (containing floating-point
+/// values) using SI units. Requires the "fp" feature.
+#[cfg(feature="fp")]
+pub fn convert_measurement_to_si(meas: Measurement) -> SIMeasurement {
+    SIMeasurement {
+        temp: (meas.temp as f32) / 1000.0,
+        temp_unit: match meas.temp_unit {
+            TempUnit::MilliDegreesCelsius => SITempUnit::Celsius,
+            TempUnit::MilliDegreesFahrenheit => SITempUnit::Fahrenheit,
+        },
+        rel_hum_percent: (meas.rel_hum_pcm as f32) / 1000.0,
+        precision: meas.precision
+    }
+}
+
+
 #[cfg(test)]
 mod tests {
     use embedded_hal_mock as hal;
-    use float_cmp::approx_eq;
 
     use self::hal::delay::MockNoop as DelayMock;
     use self::hal::i2c::{Mock as I2cMock, Transaction};
@@ -560,13 +938,13 @@ mod tests {
 
     const SHT40_I2C_ADDR: I2CAddr = I2CAddr::SHT4x_A;
 
-    fn gen_measurement_buf(temp: f32, tunit: TempUnit, rel_hum: f32) -> [u8; 6] {
-        let rel_hum_ticks = ((rel_hum + 6.0) / 125.0 * 65535.0) as u16;
+    fn gen_measurement_buf(temp: i32, tunit: TempUnit, rel_hum_pcm: u32) -> [u8; 6] {
+        let rel_hum_ticks = (((rel_hum_pcm + 6000) << 13) / 15625) as u16;
         let rel_temp_ticks;
-        if tunit == TempUnit::Celsius {
-            rel_temp_ticks = ((temp + 45.0) / 175.0 * 65535.0) as u16;
+        if tunit == TempUnit::MilliDegreesCelsius {
+            rel_temp_ticks = (((temp + 45000) << 13 ) / 21875 ) as u16;
         } else {
-            rel_temp_ticks = ((temp + 49.0) / 315.0 * 65535.0) as u16;
+            rel_temp_ticks = (((temp + 49000) << 14) / 78751 ) as u16;
         }
         let temp_buf = rel_temp_ticks.to_be_bytes();
         let temp_crc = crc8::calculate(&temp_buf);
@@ -575,12 +953,11 @@ mod tests {
         [temp_buf[0], temp_buf[1], temp_crc, rel_hum_buf[0], rel_hum_buf[1], rel_crc]
     }
 
-    fn gen_serial_buf(serial_a: u16, serial_b: u16) -> [u8; 6] {
-        let sa_buf = serial_a.to_be_bytes();
-        let sa_crc = crc8::calculate(&sa_buf);
-        let sb_buf = serial_b.to_be_bytes();
-        let sb_crc = crc8::calculate(&sb_buf);
-        [sa_buf[0], sa_buf[1], sa_crc, sb_buf[0], sb_buf[1], sb_crc]
+    fn gen_serial_buf(serial_mock: u32) -> [u8; 6] {
+        let s_buf = serial_mock.to_be_bytes();
+        let sa_crc = crc8::calculate(&s_buf[0..2]);
+        let sb_crc = crc8::calculate(&s_buf[2..4]);
+        [s_buf[0], s_buf[1], sa_crc, s_buf[2], s_buf[3], sb_crc]
     }
 
     fn gen_mock_expectations(command_expectations: &[(Command, [u8; 6])]) -> Vec<Transaction> {
@@ -596,30 +973,28 @@ mod tests {
 
     #[test]
     fn test_sht40_serialnumber() {
-        let serial_a = 23100;
-        let serial_b = 25;
+        let serial_mock: u32 = 231997321;
 
         // Mock expected I2C transactions
         let command_expectations = [
-            (Command::GetSerialNumber, gen_serial_buf(serial_a, serial_b))
+            (Command::GetSerialNumber, gen_serial_buf(serial_mock))
         ];
         let i2c_expectations = gen_mock_expectations(&command_expectations);
         let i2c_mock = I2cMock::new(&i2c_expectations);
 
         let mut sht40 = SHT40Driver::new(i2c_mock, SHT40_I2C_ADDR, DelayMock);
 
-        if let Ok(SerialNumber(a, b)) = sht40.get_serial() {
-           assert_eq!(a, serial_a);
-           assert_eq!(b, serial_b);
+        if let Ok(serial) = sht40.get_serial() {
+           assert_eq!(serial, serial_mock);
         }
     }
 
     #[test]
     fn test_sht40_measurement() {
-        let temp: f32 = 23.8;
-        let tunit = TempUnit::Celsius;
-        let rel_hum: f32 = 65.2;
-        let buf = gen_measurement_buf(temp, tunit, rel_hum); 
+        let temp: i32 = 20625;
+        let tunit = TempUnit::MilliDegreesCelsius;
+        let rel_hum_pcm: u32 = 9625;
+        let buf = gen_measurement_buf(temp, tunit, rel_hum_pcm); 
 
         // Mock expected I2C transactions
         let command_expectations = [
@@ -632,19 +1007,62 @@ mod tests {
 
         if let Ok(m) = sht40.get_temp_and_rh(Precision::High, tunit) {
            assert_eq!(m.temp_unit, tunit);
-           assert!(approx_eq!(f32, m.temp, temp, epsilon = 0.01));
-           assert!(approx_eq!(f32, m.rel_hum_percent, rel_hum, epsilon = 0.01));
+           assert_eq!(m.temp, temp);
+           assert_eq!(m.rel_hum_pcm, rel_hum_pcm);
         }
     }
     
+    #[test]
+    fn test_sht40_measurement_with_offset_conversion() {
+        let cmp_eps = 5;
+        let temp_f: i32 = 77000;
+        let temp_c: i32 = 25000;
+        let tfunit = TempUnit::MilliDegreesFahrenheit;
+        let tcunit = TempUnit::MilliDegreesCelsius;
+        let rel_hum_pcm: u32 = 9625;
+        let buf = gen_measurement_buf(temp_f, tfunit, rel_hum_pcm); 
+        let buf_2 = gen_measurement_buf(temp_c, tcunit, rel_hum_pcm); 
+
+        // Mock expected I2C transactions
+        let command_expectations = [
+            (Command::MeasureTempAndHumidity(Precision::High), buf),
+            (Command::MeasureTempAndHumidity(Precision::High), buf_2)
+        ];
+        let i2c_expectations = gen_mock_expectations(&command_expectations);
+        let i2c_mock = I2cMock::new(&i2c_expectations);
+
+        let mut sht40 = SHT40Driver::new(i2c_mock, SHT40_I2C_ADDR, DelayMock);
+        let mut t1_offset:i16 = -5000;
+        sht40.set_temp_offset(TempOffset(t1_offset, TempUnit::MilliDegreesCelsius)).unwrap();
+        let h1_offset:i16 = -1200;
+        sht40.set_hum_offset(HumPcmOffset(h1_offset)).unwrap();
+
+        if let Ok(m) = sht40.get_temp_and_rh(Precision::High, tfunit) {
+           assert_eq!(m.temp_unit, tfunit);
+           assert!((m.temp >= 68000 - cmp_eps) && (m.temp <= 68000 + cmp_eps));
+           assert_eq!(m.rel_hum_pcm, rel_hum_pcm - 1200);
+           sht40.clear_temp_offset();
+           sht40.clear_hum_offset();
+        }
+        
+        t1_offset = -9000;
+        sht40.set_temp_offset(TempOffset(t1_offset, TempUnit::MilliDegreesFahrenheit)).unwrap();
+        if let Ok(m) = sht40.get_temp_and_rh(Precision::High, tcunit) {
+           assert_eq!(m.temp_unit, tcunit);
+           assert!((m.temp >= 20000 - cmp_eps) && (m.temp <= 20000 + cmp_eps));
+           assert_eq!(m.rel_hum_pcm, rel_hum_pcm);
+        }
+
+    }
+    
     #[test]
     fn test_sht40_heater_and_measurement() {
-        let temp: f32 = 23.8;
-        let tunit = TempUnit::Celsius;
-        let rel_hum: f32 = 65.2;
+        let temp: i32 = 20625;
+        let tunit = TempUnit::MilliDegreesCelsius;
+        let rel_hum_pcm: u32 = 9625;
         let hpow = HeaterPower::High;
         let hdur = HeaterDuration::PulseShort;
-        let buf = gen_measurement_buf(temp, tunit, rel_hum); 
+        let buf = gen_measurement_buf(temp, tunit, rel_hum_pcm); 
 
         // Mock expected I2C transactions
         let command_expectations = [
@@ -658,8 +1076,37 @@ mod tests {
         if let Ok(m) = sht40.set_heater_then_measure(hpow, hdur, tunit) {
            assert_eq!(m.temp_unit, tunit);
            assert_eq!(m.precision, Precision::High);
-           assert!(approx_eq!(f32, m.temp, temp, epsilon = 0.01));
-           assert!(approx_eq!(f32, m.rel_hum_percent, rel_hum, epsilon = 0.01));
+           assert_eq!(m.temp, temp);
+           assert_eq!(m.rel_hum_pcm, rel_hum_pcm);
+        }
+    }
+
+    #[cfg(feature="fp")]
+    #[test]
+    fn test_sht40_si_measurement() {
+        let temp_si: f32 = 20.625;
+        let tunit_si = SITempUnit::Celsius;
+        let rel_hum_si: f32 = 9.625;
+
+        let temp: i32 = 20625;
+        let tunit = TempUnit::MilliDegreesCelsius;
+        let rel_hum_pcm: u32 = 9625;
+        let buf = gen_measurement_buf(temp, tunit, rel_hum_pcm); 
+
+        // Mock expected I2C transactions
+        let command_expectations = [
+            (Command::MeasureTempAndHumidity(Precision::High), buf)
+        ];
+        let i2c_expectations = gen_mock_expectations(&command_expectations);
+        let i2c_mock = I2cMock::new(&i2c_expectations);
+
+        let mut sht40 = SHT40Driver::new(i2c_mock, SHT40_I2C_ADDR, DelayMock);
+
+        if let Ok(m) = sht40.get_temp_and_rh(Precision::High, tunit) {
+           let m_si = convert_measurement_to_si(m);
+           assert_eq!(m_si.temp_unit, tunit_si);
+           assert_eq!(m_si.temp, temp_si);
+           assert_eq!(m_si.rel_hum_percent, rel_hum_si);
         }
     }
 }
